#include "OmSourceBackgroundDlg.h"
#include "OmSourcePhotometryDlg.h"

#include "OmFastPhotometryDlg.h"

#include <qapplication.h>
#include <qpushbutton.h>
#include <qdialog.h>
#include <qtable.h>
#include <qstringlist.h>
#include <qgroupbox.h>
#include <qlayout.h>
#include <qmessagebox.h>
#include <qtimer.h>
#include <qspinbox.h>
#include <qtooltip.h>
#include "OmSourceVersionNo.h"
#include "OmSourceParameters.h"
#include <iostream>
#include "OmSourceButton.h"
#include "OmSourceGroupBox.h"
#include "OmSourceLabel.h"
#include "OmSourceNumerics.h"
#include "OmImageDisplay.h"
#include "OmSourceRegions.h"
#include "OmSpinBox.h"
//#include "qapplication.h"
#include "average.h"
#include <cmath>

F90MODPROCPROTO(void,omsource_glue, deletebackgroundlist)();
F90MODPROCPROTO(void, omsource_glue, deletebackgroundregions)();



F90MODPROCPROTO(void, omsource_glue, addbackgroundregions)(OmSourceRegions *,const float &);
F90MODPROCPROTO(void, omsource_glue, getthecounts)();
F90MODPROCPROTO(void, omsource_glue, appendtobackgroundregions)(OmSourceRegions *);
F90MODPROCPROTO(float, omsource_glue, getimagecounts)(float &, float&, float&);


extern F90MODVARDEF(OmSourceParameters, omsource_glue, mybackgroundsources);
extern F90MODVARDEF(OmSourceParameters, omsource_glue, mysources);
extern F90MODVARDEF(OmSourceRegions, omsource_glue, mybackgroundregions);

extern F90MODVARDEF(int, omsource_glue, myimagenx);
extern F90MODVARDEF(int, omsource_glue, myimageny);



/********************************************************************************
 * Dialog for obtaining background regions from the user          
 *
 *******************************************************************************/

vector<OmSourceParameter> BackgroundDlg::_backgroundList; // Stores the list of background "sources"

BackgroundDlg::BackgroundDlg( QWidget *parent,  
                              OmImageDisplay *imageDisplay,
                              OmSourceParameter *source, 
                              int zoomFactor) :
    QDialog(parent, ""), _imageDisplay(imageDisplay),
    _source(source), _zoomFactor(zoomFactor), 
    _apertureRadius(6.0)

{


  //    cout << "omsourcebackgrounddlg.cc line 69, constructor, size backgroundList=" << _backgroundList.size() << endl;



    _timer = 0;
    //    _photometryDlg = dynamic_cast<PhotometryDlg*>(parent); 


    connect( this , SIGNAL(updateDisplay()), parent, SLOT ( updateDisplay()) );



  //    setCaption("PLEASE SELECT BACKROUND REGIONS BY CLICKING LEFT-MOUSE BUTTON OUTSIDE THE PHOTOMETRY ANNULI"); 
    // Create the table for storing details about the selected background regions 


    QString text("Please select background regions by\n"
                 "clicking the left mouse-button\n"
                 "The mean counts/pixel of the selected\n"
                 "regions (shown below) will be used to\n"
                 "compute the background counts of the source");
    _label = new OmSourceLabel(this, text, 10, 30, 350, 150);

    updateTimeChanged( 3 );


    _nx = (float)(F90MODPROC( omsource_glue, myimagenx ) + 1);
    _ny = (float)(F90MODPROC( omsource_glue, myimageny ) + 1);


    int xPos = 10;
    int yPos = _label->geometry().bottom() + 20;
    int height = 30;

    _averageTxt = new OmSourceLabel(this, "Mean counts/pixel=0+/-0 (n=0)",  xPos, yPos, height);
  
  
    _table = new QTable(10, 3, this, 0);
    yPos = _averageTxt->geometry().bottom() + 20;
  
    _table->setGeometry( _averageTxt->geometry().left(), yPos, 350, 120);
    yPos = _table->geometry().bottom() + 20;
  

    int noColumns = 4;


    _table->setReadOnly(true);   // Can only read the table
    _table->setNumCols( noColumns);
    _table->setNumRows(1);   
   

    QHeader *th = _table->horizontalHeader();
   
     
    th->setLabel(0,"x coord");
    _table->setText(0, 0, "x coord");
 

    th->setLabel(1,"y coord");
    _table->setText(0, 1, "y coord");
 
    th->setLabel(2,"radius");
    _table->setText(0, 2, "radius");
 

    th->setLabel(3,"counts");
    _table->setText(0, 3, "1256889999.99");


    /********************************************
    * Adjust the size of the cells
    *******************************************/

    QFont f = font();
      
    QFontMetrics fm( f );
 


    int maxWidth = 0;
    for (int i = 0; i < noColumns; i++)
    {       
        int width = fm.width( _table->text(0, i)) + 8;     
        _table->setColumnWidth(i, width  );
	_table->setText(0, i, " ");
        maxWidth += width;
    }
    maxWidth += 40;
    _table->setNumRows(0);   
    _table->resize( maxWidth, 120 );
  
    maxWidth = max( maxWidth, _table->geometry().right() );
     
    maxWidth = max( maxWidth, _label->geometry().right() );
 
    _deleteListBtn = new OmSourceButton(this, "Delete",
               "Click here to delete the source list");
  
    _deleteListBtn->setGeometry( _table->geometry().left(), _table->geometry().bottom() + 20, 100, 30  );
  
    connect(_deleteListBtn, SIGNAL(clicked()), this, SLOT (deleteList()) );


     /***********************************************************
     * Spin box to get the radius of the background aperture
     **********************************************************/
 
    yPos = _deleteListBtn->geometry().bottom() + 30;
  
   
    _apertureTxt = new OmSourceLabel( this, "Aperture",  _table->geometry().left(), yPos, height);
 
 
    xPos = _apertureTxt->geometry().right(); 
  
    _apertureSpinBox = new QSpinBox(this, "SpinBox2" );
    _apertureSpinBox->setGeometry( xPos, yPos, 50, height ); 
    
    _apertureSpinBox->setMaxValue(30);
    _apertureSpinBox->setMinValue(3);
    _apertureSpinBox->setValue((int)_apertureRadius);


    maxWidth = max(maxWidth, _apertureSpinBox->geometry().right());

    connect(_apertureSpinBox, SIGNAL(valueChanged(int)), this, SLOT(apertureRadiusHasChanged(int)));
   
    text.sprintf("Change the background aperture radius\n (allowed values=%d-%d)", 
		 3, 30);
    QToolTip::add(_apertureSpinBox, text);

    yPos = _apertureTxt->geometry().bottom() + 30;






    OmSourceGroupBox *colourBox = new OmSourceGroupBox( this,  "Ds9 image display", _apertureTxt->geometry().left(),
            _apertureTxt->geometry().bottom() + 20, 180, 70);




 
    /******************************************************************
    * A spin box to set the new regions colour
    *****************************************************************/
    
     OmSourceLabel *label1 = new OmSourceLabel( colourBox, "Background regions", 10, 30);


    _spinBoxColour = new OmSpinBox( colourBox, "Change the colour of the new background regions", OmSourceBrowser::_colourList.size() - 1, 
                     OmSourceBrowser::_colourList, OmSourceBrowser::_backgroundRegionsNo);
    _spinBoxColour->setGeometry( label1->geometry().left(), label1->geometry().bottom(), 70, 25);
    connect(_spinBoxColour, SIGNAL(valueChanged(int)), this, SLOT ( backgroundRegionColourChanged(int)));

 

    _spinBoxColour->show();
  
 

    /******************************************************************
    * A spin box to set image update time
    *****************************************************************/
    
     OmSourceLabel *label3 = new OmSourceLabel( colourBox, "update time", label1->geometry().right() + 30, label1->geometry().top());


    _spinBoxUpdate = new QSpinBox( colourBox, "Change the ds9 image update time (secs)");
    _spinBoxUpdate->setGeometry( label3->geometry().left(), label3->geometry().bottom(), 70, 25);
    connect(_spinBoxUpdate, SIGNAL(valueChanged(int)), this, SLOT ( updateTimeChanged(int)));

    QToolTip::add(_spinBoxUpdate, "The time interval (secs) at which the ds9 image will be updated with the new regions");
 
    
    _spinBoxUpdate->show();
  
    _spinBoxUpdate->setMinValue( 1 );
    _spinBoxUpdate->setMaxValue( 30 );
    
    _spinBoxUpdate->setValue( 3 );
   


    colourBox->resize( _spinBoxUpdate->geometry().right() + 20, _spinBoxUpdate->geometry().bottom() + 20 );

    maxWidth = max( maxWidth, colourBox->geometry().right() );

    yPos = colourBox->geometry().bottom() + 20;

    OmSourceButton* okButton = new OmSourceButton(this, "OK", 
         "Click here when you have selected your background regions\n"
         "The average of the mean counts/pixels for each region\n"

	 "will be used in the background-subtraction");
    okButton->setGeometry(100, yPos, 100, 30);
    connect(okButton, SIGNAL( clicked() ), this, SLOT( accept() ));

    maxWidth = max(maxWidth, okButton->geometry().right());

 
    _average = new Average;
 

    setGeometry( 0, 0, maxWidth + 10, okButton->geometry().bottom() + 20);

    //    _photometryDlg->updateDisplay();

    //   emit updateDisplay( false );

    OmSourceRegions *points1 = OmImageRegions::GetNewRegions( _imageDisplay, _nx, _ny );
    if (points1) delete points1;	  




    update();

  
    _timer = new QTimer( this);

    connect( _timer, SIGNAL(timeout()), this, SLOT(updateRegions()) );
  
    _timer->start( 2000); // 2 seconds single-shot timer
      
      
};
/***************************************
 * Destructor
 **************************************/
BackgroundDlg::~BackgroundDlg()
{

    _timer->stop();
    delete _timer;
    delete _average;
}

/**********************************************
 *
 *********************************************/
void BackgroundDlg::deleteList()
{
    _timer->stop();



   // Reset the table  
    _table->setNumRows(0);   
  
    _deleteListBtn->setDisabled(TRUE);
  

    _backgroundList.clear();
   

    _averageTxt -> setText(" ");

    _average->reset();
    



    /****************************************************
    *
    ****************************************************/


    F90MODPROC(omsource_glue, deletebackgroundlist)();
    F90MODPROC(omsource_glue, deletebackgroundregions)();







    //    _photometryDlg->updateDisplay();


    emit updateDisplay(  );
    _timer->start( 2000);
 



     
}
/************************************************************
*
*
***********************************************************/
//void BackgroundDlg::mousePressEvent( QMouseEvent *e )
//{
//    QApplication::beep();
//    if ( e->button() == QMouseEvent::LeftButton )               // left button pressed
//    QMessageBox::critical( 0, "omsource",
//			       QString("LEFT MOUSE BUTTON")); 

//}
/*********************************************************
* Checks the image for any new background regions
* and, if necessary, updates the image and table
********************************************************/
void BackgroundDlg::updateRegions()
{ 

    _timer->stop();
 
    // Get any new regions


    OmSourceRegions *newRegions = OmImageRegions::GetNewRegions(_imageDisplay, _nx, _ny);

 
    if (newRegions == 0)
    { 


        _timer->start( 2000 );

	return;
    }   

   

    // Compute the counts for each newly selected region
    float standardDeviation;
 
    //    cout << "OmSourceBackgroundDlg.cc line 401- displaying the new regions\n";

    //   cout << "size backgroundList=" << _backgroundList.size() << endl;
       
    for (int i = newRegions->lbound(); i <= newRegions->ubound(); i++)
    {   
  

        float x = (*newRegions)[i].x();
        float y = (*newRegions)[i].y();

	//        cout << i << ") x=" << x << " y=" << y << " updating the table , num rows=" << _table->numRows() << endl;
   
        float counts = F90MODPROC(omsource_glue, getimagecounts)(x, y, _apertureRadius);
	      
        // Update the table
        updateTable(x, y, _apertureRadius, counts);  


        OmSourceParameter source;
        source.xcoord = x;
        source.ycoord = y;
        source.semia = _apertureRadius;
        source.semib = _apertureRadius;
        source.photometryradius = _apertureRadius;
        source.theta = 0.0;

        _backgroundList.push_back( source );


        counts /= M_PI * _apertureRadius * _apertureRadius;
        
        float countsError = sqrt(counts);
       	
	_backgroundMode = _average->UpdateAverage(counts, countsError, standardDeviation);
    }




 
    QString text;

  
    text.sprintf("Mean counts/pixel=%7.1f+/-%7.1f (n=%d)", 
            _backgroundMode, standardDeviation, (int)_backgroundList.size() ); 
   
    //    cout << "Line 448 text=" << text << endl;



    delete( newRegions );  // Finished with the new regions
    newRegions = 0;

 



    _averageTxt -> setText(text);
    _deleteListBtn->setEnabled(TRUE);


    repaint();

    //    QApplication::processEvents();
    //    _photometryDlg->updateDisplay();

   emit updateDisplay(  );


   _table->update();


   
   //    cout << "Line 465 size backgroundList=" << _backgroundList.size() << " number of rows in table=" << _table->numRows() << endl;



    _timer->start( 2000 );

}

/*****************************************************
 * Called when the value of apertureradius 
 *spin-box has changes
 ****************************************************/
void BackgroundDlg::apertureRadiusHasChanged(int radius)
{
    _apertureRadius = radius;
    _apertureSpinBox->setValue(radius);



}
/***********************************************************
*
*
***********************************************************/
void BackgroundDlg::updateTable(float x, float y, float radius, 
           float counts)
{

        

    int rowNo = _table->numRows();
    
    //    cout << "updatetable, number of rows=" << rowNo << endl;    

    _table->setNumRows(rowNo + 1);
    
    QString text;
                 
    text.sprintf("%6.2f", x);
    _table->setText(rowNo, 0, text);

    text.sprintf("%6.2f", y);
    _table->setText(rowNo, 1, text);

    text.sprintf("%2d",  (int)radius);
    _table->setText(rowNo, 2, text);

    text.sprintf("%7.1f", counts);  
    _table->setText(rowNo, 3, text);

    _table->showRow(rowNo + 1); 

    
    //   cout << "updatetable after updating , number of rows=" << _table->numRows() << endl;    


 

}

/********************************************************
*
********************************************************/
void BackgroundDlg::update()
{
   

  //  cout << "BackgroundDlg line 527, backgroundList.size=" << _backgroundList.size() << endl;
  //  cout << "Number of rows in table=" << _table->numRows() << endl;
    if ( _backgroundList.size() == 0) return;
  

    // Compute the counts for each newly selected region

    _average->reset();

    float standardDeviation;
     
    for ( unsigned i = 0; i < _backgroundList.size(); i++)
    {   
 

        float x = _backgroundList[i].xcoord;
        float y =  _backgroundList[i].ycoord;
        float radius = _backgroundList[i].photometryradius;   
        float counts = F90MODPROC(omsource_glue, getimagecounts)(x, y, radius);
     

          // Update the table
        updateTable(x, y, radius, counts);  
	

        counts /=  M_PI * radius * radius;  
        float countsError = sqrt(counts);


	_backgroundMode = _average->UpdateAverage(counts, countsError, standardDeviation);
    }

  
   
    QString text;
    text.sprintf("Mean counts/pixel=%5.1f+/-%5.1f (n=%d)", 
            _backgroundMode, standardDeviation, (int)_backgroundList.size() ); 
   
    _averageTxt -> setText(text);

    _deleteListBtn->setEnabled(TRUE);


} 


/****************************************************************************
*
*
****************************************************************************/
void BackgroundDlg::updateTimeChanged( int time )
{

    QString text1;
    text1.sprintf("%2d", time);
    QString text=QString("Please select new background regions by clicking\n")   + 
                 QString("the left mouse-button. The regions can be deleted\n")  +
                 QString("by clicking the 'Delete' button. The selected\n")      + 
                 QString("sources will be displayed in ")                        +
                 QString( OmSourceBrowser::_backgroundRegionsColour.c_str())     +
                 QString(".\nThe DS9 image will be updated every ")              +
                 text1 + QString("\n")                                           +
                 QString("seconds, approximately.\n")                            +
                 QString("The mean counts/pixel computed from the regions\n")    +
                 QString("will be used to subtract the background.");


    _label->setText1( text);

    _time = time * 1000;
   

    if ( _timer)
    {
        _timer->stop();
        _timer->start( _time );
    }
  
}

/****************************************************************************
*
*
****************************************************************************/
void BackgroundDlg::backgroundRegionColourChanged( int no )
{

 
    OmSourceBrowser::_backgroundRegionsColour = OmSourceBrowser::_colourList[no - 1];

    OmSourceBrowser::_backgroundRegionsNo = no;

    //    _photometryDlg->updateDisplay();

    emit updateDisplay(  );



}















