!******************************************
! Creates a test image
!*****************************************
module omskybackgroundimage_mod

use dal
use cal
use omskybackground_mod
use omutility_mod
use omsmoothimage_mod
use omsourceparam_mod
use ominterpolate_mod
use cubicshepard_mod
use gridinterpolation_mod
use omprogress_mod
implicit none



interface createBackgroundImage
    module procedure  createBackgroundImage1,  &
                      createBackgroundImage2,  &
                      createBackgroundImage3,  &
                      createBackgroundImage4,  &
                      createBackgroundImage5,  &
                      createBackgroundImage6

end interface



!****************************************
! A structure to store pixel information
!****************************************
Type PixelStruct
    integer(int16) :: x = 0
    integer(int16) :: y = 0
    integer(int16) :: number = 0
    real(single) :: counts = 0.0
    real(single) :: backgroundCounts = 0.0
    integer(int16) :: noNeighbours = 0
end Type PixelStruct




contains


!******************************************************
!
!******************************************************
subroutine createBackgroundImage1(image, qualityImage, &
                                  backgroundimage, backgroundsigmaimage)
    real(single), pointer :: image(:, :)
    logical(boolean), pointer :: qualityImage(:, :)
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
  


    call createBackgroundImage2(image, qualityImage, 10000_int16, &
                                backgroundimage=backgroundimage, backgroundsigmaimage=backgroundsigmaImage)

end subroutine createBackgroundImage1


!***************************************************************
!
!**************************************************************
subroutine createBackgroundImage2(image, qualityImage, width, &
               backgroundImage, backgroundsigmaImage, maxheight)
    real(single), pointer :: image(:, :)
    logical(boolean), pointer :: qualityImage(:, :)
 
  
    integer(int16), intent(in) :: width
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    real(single), intent(in), optional :: maxHeight

    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
   
    real(single) :: globalMedian, sigma, mode
    integer(int16), pointer :: levelImage(:, :) => null()
    logical, pointer :: pixelOkay(:, :) => null()
  
    integer(int16) :: dw = 50  
    integer(int16) :: dw1
    type(OmProgressT) :: reporter   

    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)

    nx = ubound(image, 1)
    ny = ubound(image, 2)
 
    
    dr = min(width, nx)
    dr = min(dr, ny)
    dx = dr / 2
    dy = dx 
 
    dr = 2 * dx  

    dw = 50
    dw1 = dw - 1
    

    levelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "levelimage")

    mode = ComputeBackgroundMode4(image, levelImage, qualityImage, &
               nx + 1_int16, ny + 1_int16, 2.0, sigma)

 
    if (present(maxHeight)) then
        globalMedian = mode + maxheight * max(1.0, sigma)
    else
        globalMedian = mode + 10.0 * max(1.0, sigma)
    end if

    deallocate(LevelImage)
    nullify(levelImage)
     
    if (.not. associated(backgroundimage)) then
    

        backgroundImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if

    if (.not. associated(backgroundsigmaimage)) then
 
        backgroundSigmaImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if




    backgroundImage = image


    allocate(pixelOkay(0_int16 : nx, 0_int16 : ny))
    where (image .lt. globalMedian .and. .not. qualityImage)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 
  
    globalMedian = mode - 5.0 * max(1.0, sigma)

    where(image .lt. globalMedian) pixelOkay = .false.


    call construct(reporter, label="Computing background image", steps = 100 * int(nx / dw))


    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    do i = 0, nx, dw
          iu = min(i + dw1, nx)
        do j = 0, ny, dw
            ju = min(j + dw1, ny)


!            write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Creating background  strip ", i, ":", iu, ",", j, ":", ju
!            call message(AppMsg, Verbose)

            
            call CreateBackgroundStrip(i, iu, j, ju, image, &
               backgroundImage, backgroundsigmaImage, pixelOkay, &
               dx, dy, dr, mode)

        end do

        call setProgress(reporter, 100 * int(i / dw))

    end do

   call destruct(reporter)
   
    deallocate(pixelOkay)
    nullify(pixelOkay)




    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  

    call smoothimage( backgroundImage, int(nx + 1, int32), int(ny + 1, int32), &
          int(20 / binx, int32), int(20 / binx, int32))

    
!    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fit", "background image")
!call fatal("stopped displaying modeimage.fit")  
!    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

end subroutine createBackgroundImage2

!***************************************************************
!
!**************************************************************
subroutine createBackgroundImage5(image, width, maxheight)
    real(single), pointer :: image(:, :)
   
    integer(int16), intent(in) :: width
    real(single), intent(in), optional :: maxHeight

    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
   
    real(single) :: globalMedian, sigma, mode
 
    logical, pointer :: pixelOkay(:, :) => null()
    real(single), pointer :: tempImage1(:, :) => null()
    real(single), pointer :: tempImage2(:, :) => null()

  
    integer(int16) :: dw = 50  
    integer(int16) :: dw1
   
    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    dr = min(width, nx)
    dr = min(dr, ny)
    dx = dr / 2
    dy = dx 
 
    dr = 2 * dx  

    dw = max(dr + 10, 50)
    dw = 50
    dw1 = dw - 1
    


    tempImage1 => allocate2dSingle(0_int16, nx, 0_int16, ny, "tempimage1")
    tempImage2 => allocate2dSingle(0_int16, nx, 0_int16, ny, "tempimage2")

    tempimage1 = image

    mode = ComputeBackgroundMode2(image, nx + 1_int16, ny + 1_int16, 2.0, sigma)

 
    if (present(maxHeight)) then
        globalMedian = mode + maxheight * max(1.0, sigma)
    else
        globalMedian = mode + 3.0 * max(1.0, sigma)
    end if
   
     

    allocate(pixelOkay(0_int16 : nx, 0_int16 : ny))
    where (image .lt. globalMedian)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 
  
   

  

    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    do i = 0, nx, dw

        iu = min(i + dw1, nx)
        do j = 0, ny, dw
            ju = min(j + dw1, ny)


!            write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Creating background  strip ", i, ":", iu, ",", j, ":", ju
!            call message(AppMsg, Verbose)

            
            call CreateBackgroundStrip(i, iu, j, ju, image, &
               tempimage1, tempimage2, pixelOkay, &
               dx, dy, dr, mode)

        end do


    end do


    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  


    image = tempimage1
    deallocate(pixelOkay, tempimage1, tempimage2)
    nullify(pixelOkay, tempimage1, tempimage2)

    call smoothimage(image, int(nx + 1, int32), int(ny + 1, int32), &
          int(20 / binx, int32), int(20 / binx, int32))



!    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fit", "background image")
!call fatal("stopped displaying modeimage.fit")  
!    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

end subroutine createBackgroundImage5

!***************************************************************
!
!**************************************************************
subroutine createBackgroundImage6(image, levelImage, width, maxheight)
    real(single), pointer :: image(:, :)
    integer(int16), pointer :: levelImage(:, :)

    integer(int16), intent(in) :: width
    real(single), intent(in), optional :: maxHeight

    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
   
    real(single) :: globalMedian, sigma, mode
  
    logical, pointer :: pixelOkay(:, :) => null()
    real(single), pointer :: tempImage1(:, :) => null()
    real(single), pointer :: tempImage2(:, :) => null()

  
    integer(int16) :: dw = 50  
    integer(int16) :: dw1
   
    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    dr = min(width, nx)
    dr = min(dr, ny)
    dx = dr / 2
    dy = dx 
 
    dr = 2 * dx  

    dw = max(dr + 10, 50)
    dw = 50
    dw1 = dw - 1
    


    tempImage1 => allocate2dSingle(0_int16, nx, 0_int16, ny, "tempimage1")
    tempImage2 => allocate2dSingle(0_int16, nx, 0_int16, ny, "tempimage2")

    tempimage1 = image

    mode = ComputeBackgroundMode(image, levelimage, nx + 1_int16, ny + 1_int16, 2.0, sigma)

 
    if (present(maxHeight)) then
        globalMedian = mode + maxheight * max(1.0, sigma)
    else
        globalMedian = mode + 3.0 * max(1.0, sigma)
    end if
    
     

    allocate(pixelOkay(0_int16 : nx, 0_int16 : ny))
    where (image .lt. globalMedian .and. levelimage .eq. 0_int16)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 
  
   
    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    do i = 0, nx, dw

        iu = min(i + dw1, nx)
        do j = 0, ny, dw
            ju = min(j + dw1, ny)


!            write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Creating background  strip ", i, ":", iu, ",", j, ":", ju
!            call message(AppMsg, Verbose)

            
            call CreateBackgroundStrip(i, iu, j, ju, image, &
               tempimage1, tempimage2, pixelOkay, &
               dx, dy, dr, mode)

        end do


    end do


    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  


    image = tempimage1
    deallocate(pixelOkay, tempimage1, tempimage2)
    nullify(pixelOkay, tempimage1, tempimage2)

    call smoothimage(image, int(nx + 1, int32), int(ny + 1, int32), &
          int(20 / binx, int32), int(20 / binx, int32))



!    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fit", "background image")
!call fatal("stopped displaying modeimage.fit")  
!    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

end subroutine createBackgroundImage6



!***************************************************************
!
!**************************************************************
subroutine createBackgroundImage3(image, width, &
               backgroundImage, backgroundsigmaImage, maxheight)
    real(single), pointer :: image(:, :)
  
  
    integer(int16), intent(in) :: width
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    real(single), intent(in), optional :: maxHeight

    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
   
    real(single) :: globalMedian, sigma, mode
    integer(int16), pointer :: levelImage(:, :) => null()
    logical, pointer :: pixelOkay(:, :) => null()
    logical(boolean), pointer :: qualityImage(:, :)
 
 
    integer(int16) :: dw = 50  
    integer(int16) :: dw1
  
   
    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    allocate(qualityImage(0 : nx, 0 : ny))

    dr = min(width, nx)
    dr = min(dr, ny)
    dx = dr / 2
    dy = dx 
 
    dr = 2 * dx  

    dw = max(dr + 10, 50)
    dw1 = dw - 1
 

    where(image .lt. 0.0) 
        qualityImage = .false.
    elsewhere
        qualityImage = .true.
    end where

    levelImage => allocate2dInt16(0_int16, nx, 0_int16, ny, "levelimage")
    mode = ComputeBackgroundMode4(image, levelImage, qualityImage, &
               nx + 1_int16, ny + 1_int16, 2.0, sigma)

 
    if (present(maxHeight)) then
        globalMedian = mode + maxheight * max(1.0, sigma)
    else
        globalMedian = mode + 10.0 * max(1.0, sigma)
    end if
    deallocate(LevelImage)
    nullify(levelImage)



     
    if (associated(backgroundimage)) deallocate(backgroundimage)
    if (associated(backgroundSigmaimage)) deallocate(backgroundSigmaimage)

    backgroundImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    backgroundSigmaImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")



    backgroundImage = image


    allocate(pixelOkay(0 : nx, 0 : ny))
    where (image .lt. globalMedian .and. .not. qualityImage)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 
    globalMedian = mode - 5.0 * max(1.0, sigma)

    where(image .lt. globalMedian) pixelOkay = .false.



    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    do i = 0, nx, dw

        iu = min(i + dw1, nx)
        do j = 0, ny, dw
            ju = min(j + dw1, ny)


!            write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Creating background  strip ", i, ":", iu, ",", j, ":", ju
!            call message(AppMsg, Verbose)

            
            call CreateBackgroundStrip(i, iu, j, ju, image, &
               backgroundImage, backgroundsigmaImage, pixelOkay, &
               dx, dy, dr, mode)

        end do


    end do






    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  

  
    call smoothimage( backgroundImage, int(nx + 1, int32), int(ny + 1, int32), &
          int(20 / binx, int32), int(20 / binx, int32))

  


    deallocate(pixelOkay, qualityImage)
    nullify(pixelOkay, qualityImage)



!    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fits", "background image")
!call fatal("stopped displaying modeimage.fits")  
!    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

end subroutine createBackgroundImage3


!***************************************************************
!
!**************************************************************
subroutine createBackgroundImage4(image, qualityImage, width, &
               backgroundImage, backgroundsigmaImage,         &
               levelimage, maxheight)
    real(single), pointer :: image(:, :)
    logical(boolean), pointer :: qualityImage(:, :)
 
  
    integer(int16), intent(in) :: width
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    integer(int16), pointer :: levelImage(:, :) 
 
    real(single), intent(in), optional :: maxHeight

    integer(int16) :: nx,  ny
    integer(int16) :: i, j
    integer(int16) :: iu, ju, dx, dy, dr
   
    real(single) :: globalMedian, sigma, mode
    logical, pointer :: pixelOkay(:, :) => null()
  
    integer(int16) :: dw = 50  
    integer(int16) :: dw1
   
    write(errstr,'(a)') "Creating background  image"
    call message(AppMsg, Verbose)
    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    dr = min(width, nx)
    dr = min(dr, ny)
    dx = dr / 2
    dy = dx 
 
    dr = 2 * dx  

    dw = max(dr + 10, 50)
    dw = 50
    dw1 = dw - 1
    

    mode = ComputeBackgroundMode4(image, levelImage, qualityImage, &
               nx + 1_int16, ny + 1_int16, 2.0, sigma)

 
    if (present(maxHeight)) then
        globalMedian = mode + maxheight * max(1.0, sigma)
    else
        globalMedian = mode + 10.0 * max(1.0, sigma)
    end if
    
     
    if (.not. associated(backgroundimage)) then
    

        backgroundImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if

    if (.not. associated(backgroundsigmaimage)) then
 
        backgroundSigmaImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempImage")
    end if


    backgroundImage = image


    allocate(pixelOkay(0_int16 : nx, 0_int16 : ny))
    where (image .lt. globalMedian .and. .not. qualityImage .and. levelimage .eq. 0)
        pixelOkay = .true.
    elsewhere
        pixelOkay = .false.
    end where 
    globalMedian = mode - 5.0 * max(1.0, sigma)

    where(image .lt. globalMedian) pixelOkay = .false.


    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    do i = 0, nx, dw

        iu = min(i + dw1, nx)
        do j = 0, ny, dw
            ju = min(j + dw1, ny)


!            write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Creating background  strip ", i, ":", iu, ",", j, ":", ju
!            call message(AppMsg, Verbose)

            
            call CreateBackgroundStrip(i, iu, j, ju, image, &
               backgroundImage, backgroundsigmaImage, pixelOkay, &
               dx, dy, dr, mode)

        end do


    end do






    write(errstr,'(a)') "Smoothing background  image"
    call message(AppMsg, Verbose)
  

  
 
    call smoothimage( backgroundImage, int(nx + 1, int32), int(ny + 1, int32), &
          int(20 / binx, int32), int(20 / binx, int32))






    deallocate(pixelOkay)
    nullify(pixelOkay)



!    call saveImage(backgroundimage, nx + 1_int16, ny + 1_int16, "modeimage.fit", "background image")
!call fatal("stopped displaying modeimage.fit")  
!    call saveImage(backgroundsigmaimage, nx + 1, ny + 1, "sigmaimage.fit", "background image")
  

end subroutine createBackgroundImage4

!***************************************************************
!
!**************************************************************
subroutine CreateBackgroundStripNew(il, iu, jl, ju, image, &
               backgroundImage, backgroundsigmaImage, pixelOkay, &
               dx, dy, dr, mode)

    integer(int16), intent(in) :: il, iu, jl, ju
    real(single), pointer :: image(:, :)

  
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    logical, pointer :: pixelOkay(:, :)
    integer(int16), intent(in) :: dx, dy, dr
    real(single), intent(in) :: mode

    integer(int16) :: i, j, l, m, k
    integer(int16) :: il1, iu1, jl1, ju1, nx, ny
    integer(int32) :: count, pixelSize, n   

    logical, pointer :: usepixel(:) => null()
    real(single) :: sigma 
    real(single), pointer :: pixelValue(:) => null()
    integer, pointer :: pixelIndex(:, :) => null()
    integer(int16) :: twoDr
    integer(int16), parameter :: maxNoPixels = 1000
    !***************************************************************
    !
    !
    !**************************************************************
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    twoDr = 2_int16 * dr
    il1 = max(0_int16, il - twoDr)
    iu1 = min(nx, iu + twoDr)
    jl1 = max(0_int16, jl - twoDr)
    ju1 = min(ny, ju + twoDr)

    call SetUpIndices(il1, iu1, jl1, ju1, image, pixelOkay, pixelValue, pixelIndex)


    pixelSize = size(pixelValue)
    allocate(usePixel(1 : pixelSize))

    pixelSize = size(pixelValue)
    allocate(usePixel(1 : pixelSize))

    usePixel = .false.

!    backgroundImage(il : iu, jl : ju) = -2.0
!    backgroundsigmaImage(il : iu, jl : ju) = -2.0


    ! Create the background image
    do j = jl, ju        

        jl1 = j - dy
        ju1 = j + dy

        if (jl1 .lt. 0) then
            jl1 = 0
            ju1 = min(twoDR, ny)
           
        else if (ju1 .ge. ny) then
            ju1 = ny
            jl1 = max(0_int16, ju1 - twoDr)
                
        end if
  
       
        do i = il, iu

            if (image(i, j) .lt. 0.0) then ! Ignore negative image values

                backgroundimage(i, j) = -1.0
                backgroundsigmaimage(i, j) = 1000.0
                cycle
            end if

            il1 = i - dx
            iu1 = i + dx

            if (il1 .lt. 0) then
                il1 = 0
                iu1 = min(twoDr, nx)


            else if (iu1 .ge. nx) then
         
                iu1 = nx
                il1 = max(0_int16, iu1 - twoDr)
          
            end if
           



            !*********************************************************************
            ! Get the pixels within the box area for the background determination
            !*********************************************************************
     
            count = 0
            ! The centre pixel

            if (pixelOkay(i, j)) then
                count = count + 1                        
                n = pixelIndex(i, j)
                usePixel(n) = .true.                      
            end if
           
             
mLoop : do m = 1, 1000

              
!                write(*, *) "Line 806 i=", i, " j=", j
                ! The row of pixels below i, j
               
                l = j - m 
                if (l .ge. jl1) then
                    do k = max(il1, i - m), min(i + m, iu1)
                        if (pixelOkay(k, l)) then
                            count = count + 1                        
                            n = pixelIndex(k, l)
                            usePixel(n) = .true.                      
                        end if
                        if (count .ge. maxNoPixels) exit mLoop
                    end do
                end if
         
!            write(*, *) "Line 820 i=", i, " j=", j
         
                ! Row of pixels  above i, j               
                l = j + m 
                if (l .le. ju1) then
                    do k = max(il1, i - m), min(i + m, iu1)
                        if (pixelOkay(k, l)) then
                            count = count + 1                        
                            n = pixelIndex(k, l)
                            usePixel(n) = .true.                      
                        end if
                     end do

                     if (count .ge. maxNoPixels) exit mLoop
    
                end if     


!write(*, *) "Line 837 i=", i, " j=", j
         

                

                ! The column of pixels to the left of i, j               
                l = i - m
                if (l .ge. il1) then 
                    do k = max(jl1, j - m + 1_int16), min(j + m - 1_int16, ju1)
                        if (pixelOkay(l, k)) then
                            count = count + 1                        
                            n = pixelIndex(l, k)
                            usePixel(n) = .true.                      
                        end if
                     end do
                     if (count .ge. maxNoPixels) exit mLoop

               end if

!     write(*, *) "Line 854 i=", i, " j=", j
         

                ! The column of pixels to the right of i, j               
                l = i + m
                if (l .le. iu1) then 
                    do k = max(jl1, j - m + 1_int16), min(j + m - 1_int16, ju1)
                        if (pixelOkay(l, k)) then
                            count = count + 1                        
                            n = pixelIndex(l, k)
                            usePixel(n) = .true.                      
                        end if
                        
                    end do
                    if (count .ge. maxNoPixels) exit mLoop
    
                end if






            end do mLoop     

            if (count .gt. 10) then  ! Was 10
!write(*, *) "Line 820 count=", count


                backgroundimage(i, j) = BackgroundMode11(count, usePixel, pixelValue, &
                        0.2, 2.0, sigma) 

                backgroundsigmaimage(i, j) = max(1.0, sigma) 
           else

                backgroundimage(i, j) = mode ! image(i, j)
                backgroundsigmaimage(i, j) = max(1.0, sqrt(image(i, j)))
                usePixel = .false.

            end if



        end do
    end do



!    do j = jl, ju        
!        do i = il, iu
!            if (abs(backgroundimage(i, j) - 2.0) .lt. 0.0001) then
!                backgroundimage(i, j) = pixelAverage(i, j, backgroundimage) 
!                backgroundsigmaimage(i, j) = pixelAverage(i, j, backgroundsigmaimage) 

!            end if 

!        end do
!    end do

    deallocate(usePixel, pixelValue, pixelIndex)
    nullify(usePixel, pixelValue, pixelIndex)

 
end subroutine CreateBackgroundStripNew

!***************************************************************
!
!**************************************************************
subroutine CreateBackgroundStrip(il, iu, jl, ju, image, &
               backgroundImage, backgroundsigmaImage, pixelOkay, &
               dx, dy, dr, mode)

    integer(int16), intent(in) :: il, iu, jl, ju
    real(single), pointer :: image(:, :)

  
    real(single), pointer :: backgroundImage(:, :)
    real(single), pointer :: backgroundSigmaImage(:, :)
    logical, pointer :: pixelOkay(:, :)
    integer(int16), intent(in) :: dx, dy, dr
    real(single), intent(in) :: mode

    integer(int16) :: i, j, l, k
    integer(int16) :: il1, iu1, jl1, ju1, nx, ny
    integer(int32) :: count, pixelSize, n   

    logical, pointer :: usepixel(:) => null()
    real(single) :: sigma 
    real(single), pointer :: pixelValue(:) => null()
    integer, pointer :: pixelIndex(:, :) => null()
    integer(int16) :: twoDr

    !***************************************************************
    !
    !
    !**************************************************************
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    twoDr = 2_int16 * dr

    il1 = max(0_int16, il - twoDr)
    iu1 = min(nx, iu + twoDr)
    jl1 = max(0_int16, jl - twoDr)
    ju1 = min(ny, ju + twoDr)




!    il1 = max(0_int16, il - dx - 1_int16)
!    iu1 = min(nx, iu + dx + 1_int16)
!    jl1 = max(0_int16, jl - dy - 1_int16)
!    ju1 = min(ny, ju + dy + 1_int16)




    call SetUpIndices(il1, iu1, jl1, ju1, image, pixelOkay, pixelValue, pixelIndex)


    pixelSize = size(pixelValue)
    allocate(usePixel(1 : pixelSize))

    pixelSize = size(pixelValue)
    allocate(usePixel(1 : pixelSize))

    usePixel = .false.

!    backgroundImage(il : iu, jl : ju) = -2.0
!    backgroundsigmaImage(il : iu, jl : ju) = -2.0


    ! Create the background image
    do j = jl, ju        

        jl1 = j - dy
        ju1 = j + dy

        if (jl1 .lt. 0) then
            jl1 = 0
            ju1 = min(twoDR, ny)
           
        else if (ju1 .ge. ny) then
            ju1 = ny
            jl1 = max(0_int16, ju1 - twoDr)
                
        end if
  
       
        do i = il, iu
!            if (mod((i + 1) *(j + 1) , 2) .ne. 0) cycle
            if (image(i, j) .lt. 0.0) then ! Ignore negative image values

                backgroundimage(i, j) = -1.0
                backgroundsigmaimage(i, j) = 1000.0
                cycle
            end if
            il1 = i - dx
            iu1 = i + dx

            if (il1 .lt. 0) then
                il1 = 0
                iu1 = min(twoDr, nx)


            else if (iu1 .ge. nx) then
         
                iu1 = nx
                il1 = max(0_int16, iu1 - twoDr)
          
            end if
                
            count = 0           
     
            do l = jl1, ju1

                do k = il1, iu1

                    if (pixelOkay(k, l)) then
                        count = count + 1
                        
                        n = pixelIndex(k, l)
                        usePixel(n) = .true.
                      
                    end if
                end do
            end do     

            if (count .gt. 10) then  ! Was 10

                backgroundimage(i, j) = BackgroundMode11(count, usePixel, pixelValue, &
                        0.2, 2.0, sigma) 

                backgroundsigmaimage(i, j) = max(1.0, sigma) 
           else

                backgroundimage(i, j) = mode ! image(i, j)
                backgroundsigmaimage(i, j) = max(1.0, sqrt(image(i, j)))
                usePixel = .false.

            end if



        end do
    end do



!    do j = jl, ju        
!        do i = il, iu
!            if (abs(backgroundimage(i, j) - 2.0) .lt. 0.0001) then
!                backgroundimage(i, j) = pixelAverage(i, j, backgroundimage) 
!                backgroundsigmaimage(i, j) = pixelAverage(i, j, backgroundsigmaimage) 

!            end if 

!        end do
!    end do

    deallocate(usePixel, pixelValue, pixelIndex)
    nullify(usePixel, pixelValue, pixelIndex)

 
end subroutine CreateBackgroundStrip




!************************************************
!
!***********************************************
real(single) Function PixelAverage(i, j, image)
    integer(int16), intent(in) :: i, j
    real(single), pointer :: image(:, :)
    integer(int16) :: nx, ny, k, il, iu, jl, ju, i1, j1
    real(double) :: sum

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    
    il = max(0_int16, i - 1_int16)
    iu = min(nx, i + 1_int16)
    jl = max(0_int16, j - 1_int16)
    ju = min(ny, j + 1_int16)



    k = 0
    sum = 0.0_double
    do i1 = il, iu
        do j1 = jl, ju
            if (i1 .eq. i .and. j1 .eq. j) cycle
            if (image(i1, j1) .lt. 0.0) cycle
            k = k + 1
            sum = sum + dble(image(i1, j1))
         end do
    end do  
  

  
    if (k  .gt. 0) then
        PixelAverage = sum  / dble(k)
    else
        PixelAverage = image(i, j)
    end if

end Function PixelAverage

!*****************************************************************
!
!****************************************************************
subroutine initialiseMedian(image, pixelOkay, pixelIndex, pixelValue, usePixel)
    real(single), pointer :: image(:, :)
    logical, pointer :: pixelOkay(:, :)
    integer, pointer :: pixelIndex(:)
    real(single), pointer :: pixelValue(:)
    logical, pointer :: usePixel(:, :)
  
    integer :: i, j, k, nx, ny

    nx = ubound(image, 1)
    ny = ubound(image, 2)
    k = count(pixelOkay)
    pixelValue => allocateSingle(1, k, "pixelValue")
    pixelIndex => allocateInt32(1, k, "pixelIndex")
    allocate(usePixel(1, k))

    k = 0
    do i = 0, nx

        do j = 0, ny
            if (pixelOkay(i, j)) then
                k = k + 1
                pixelValue(k) = image(i, j) ! The list of usable pixels for the median
            end if
        end do
    end do
 
    ! Index sort them

    ! Index sort the pixel values, pixelIndex stores the pixelvalue 
    ! indices in order of increasing pixel value
    call IndexSort(pixelValue, pixelIndex)


end subroutine initialiseMedian

!********************************************************************
! subroutine SetUpIndices
! 
!
!********************************************************************

subroutine SetUpIndices(il, iu, jl, ju, image, usePixel, pixelValue, pixelIndex)
    integer(int16), intent(in) :: il, iu, jl, ju
    real(single), pointer :: image(:, :)
    logical, pointer :: usePixel(:, :)
    real(single), pointer :: pixelValue(:) 
    integer, pointer :: pixelIndex(:, :)
  
    integer(int32), pointer, dimension(:) :: arrayIndex => null()

    integer :: i, j, k
    type(PixelStruct), pointer :: pixel(:) => null()
 

  

    k = count(usePixel(il : iu, jl : ju)) ! The number of pixels to be checked
!write(*, *) "Line 698 k=", k  
    arrayIndex => allocateInt32(1, k, "arrayIndex")

    if (associated(pixelValue)) deallocate(pixelValue)
    if (associated(pixelIndex)) deallocate(pixelIndex)

    pixelValue => allocateSingle(1, k, "pixelValue")
  

    allocate(pixelIndex(il : iu, jl : ju)) 
    allocate(pixel(1 : k))

    k = 0 

    do j = jl, ju
        do i = il, iu
            if (usePixel(i, j)) then
                k = k + 1
                pixel(k)%x = i
                pixel(k)%y = j
                pixel(k)%counts = image(i, j)
 
            end if
        end do
    end do
!write(*, *) "Line 724 k=", k  


    ! Index sort the pixel values- pixelValue will be sorted into 
    ! increasing pixel value

    call IndexSort(pixel%counts, arrayIndex)


    do i = 1, k
        j = arrayIndex(i)
        pixelValue(i) = pixel(j)%counts
        pixelIndex(pixel(j)%x, pixel(j)%y) = i
    end do


    deallocate(pixel, arrayIndex)
    nullify(pixel, arrayIndex)

end subroutine SetUpIndices


!*****************************************
! Function BackgroundMode
!*****************************************
function BackgroundMode10(nData, data, &
                        tolerance, nSigma, sigma) result(mode)
    integer(int32), intent(in) :: nData
    real(single) :: data(1 : ndata)

    real(single), intent(in) :: tolerance    ! Tolerance is usually set to 0.2
    real(single), intent(in)  :: nSigma
    real(single), intent(out) :: sigma
 
    integer(int32) :: nlist
   
    real(single)    :: mode
    integer(int32) :: i, k
     
    real(single)   :: median
    real(double)   :: variance, testSigma
    real(single)   :: highValue
    integer(int32) :: m, nmin

    nList = nData

    nmin = ishft(ndata, -1)
             
    call HeapSort(data, nData)


  
    ! reject data at more than nSigma from data
   
    testSigma = 0.0
!write(*, *) "background mode 1"
    sigmaClip: do i = 1, 20   !  was 20
      
             
        median = SortedMedian(data, nlist)
               
    


!        write(*, *) "i = ", i, " median=", median, " nlist=", nlist
        !*******************************
        ! Now do sigma clipping
        ! Get variance and std deviation
        !********************************
            
        variance = dot_product(data(1 : nlist) - median, data(1 : nlist) - median) 
        sigma = sqrt(variance / dble(nlist - 1))
     
      
        highValue = median + nSigma * sigma

!        lowValue = median - nsigma * sigma

   
        m = 0 
        do k = 1, nlist
!            if(data(k) .le. highValue .and. data(k) .ge. lowValue) then
            if(data(k) .le. highValue) then

  
                m = m + 1
                data(m) = data(k)
               
            end if
        end do
       

        if(m .lt. nmin) exit sigmaClip
     
        if(abs((sigma - testSigma) / sigma) .lt. tolerance) exit sigmaClip
     

       
        nList = m
   
        testSigma = sigma
        
    end do sigmaClip

  
    
    mode = sum(data(1 : nList)) / dble(nList)
   
    

!write(*, *) "Background mode=", mode, "+/-", sigma, " n=", nlist, " no of iterations=", i  
    
end function BackgroundMode10



!*****************************************
! Function BackgroundMode
!*****************************************
function BackgroundMode11(nData, usePixel, pixelValue, &
                        tolerance, nSigma, sigma) result(mode)
    integer(int32), intent(in) :: nData

    logical :: usePixel(:)
    real(single) :: pixelValue(: )

    real(single), intent(in) :: tolerance    ! Tolerance is usually set to 0.2
    real(single), intent(in)  :: nSigma
    real(single), intent(out) :: sigma

    real(single) :: data(1 : nData) 
    integer(int32) :: nlist
   
    real(single)    :: mode
    integer(int32) :: i, k
    real(single)   :: median
    real(double)   :: variance, testSigma, diff
    real(single)   :: highValue, lowValue
    integer(int32) :: m, nmin



    k = 0

    do i = 1, size(usePixel)
        if (usePixel(i)) then
            k = k + 1

            data(k) = pixelValue(i)
        
            usePixel(i) = .false.

        end if
    end do


    nList = nData

    nmin = ishft(ndata, -1)
             
  
    ! reject data at more than nSigma from data
   
    testSigma = 0.0

    sigmaClip: do i = 1, 20   !  was 20
      
             


        !********************************************************************
        ! Get the median
        !*******************************************************************
        k = ishft(nlist, -1)
        if (ishft(k, 1) .eq. nlist) then   ! Even number
            median = (data(k) + data(k + 1)) / 2.0
        else
            median = data(k + 1)
        end if

        !*******************************
        ! Now do sigma clipping
        ! Get variance and std deviation
        !********************************
            
        variance = 0.0_double
        do k = 1, nList
            diff = data(k) - median
            variance = variance + diff * diff 
         
        end do
        sigma = sqrt(variance / dble(nlist - 1))
     
      
        highValue = median + nSigma * sigma

        lowValue = median - nsigma * sigma

   
        m = 0  
        do k = 1, nlist
            if (data(k) .lt. lowValue) cycle ! Put in on Monday
            if(data(k) .gt. highValue) exit
  
            m = m + 1
            data(m) = data(k)
        end do
       

        if(m .lt. nmin .or. sigma .lt. 0.1) exit sigmaClip
     
        if(abs((sigma - testSigma) / sigma) .lt. tolerance) exit sigmaClip
     

       
        nList = m
   
        testSigma = sigma
        
    end do sigmaClip

  
    
    mode = sum(data(1 : nList)) / dble(nList)
            
    variance = 0.0_double
    do k = 1, nList
        diff = data(k) - mode
        variance = variance + diff * diff 
         
    end do
    sigma = sqrt(variance / dble(nlist - 1))
 
    
end function BackgroundMode11





!*****************************************************************
!
! RefineBackgroundImage
! Interpolates background across extended-source regions
!
!*****************************************************************
subroutine RefineBackgroundImage(image)
    real(single), pointer :: image(:, :)
  

    integer(int16) :: i, j, nx, ny, iu, ju, dx  
    real(single), pointer :: tempImage(:, :)
   
  
    write(errstr,'(a)') "Refining background  image"
    call message(AppMsg, Verbose)


    nx = ubound(image, 1)
    ny = ubound(image, 2)

    tempImage => allocate2dsingle(0_int16, nx, 0_int16, ny, "tempimage")
    tempimage = image


    !*************************************************************
    ! Create the background image in sections of 100 x 100 pixels
    !*************************************************************  
    dx = 200 / binx
    do i = 0, nx, dx

        iu = min(i + dx, nx)
        do j = 0, ny, dx
            ju = min(j + dx, ny)


 !           write(errstr,'(a, i5, a, i5, a, i5, a, i5)') "Refining background  strip ", i, ":", iu, ",", j, ":", ju
 !           call message(AppMsg, Verbose)

            
            call RefineBackgroundStrip(i, iu, j, ju, image, tempImage)
            

        end do


    end do

!    call Kuwaharafilter(tempImage)
!    call Lpdfilter(tempImage)


!    call smoothimage( tempImage, int(nx + 1, int32), int(ny + 1, int32), &
!          int(20 / binx, int32), int(20 / binx, int32))

        
!call saveImage(image, nx + 1_int16, ny + 1_int16, "image.fits", "background image")
!call saveImage(tempimage, nx + 1_int16, ny + 1_int16, "tempimage.fits", "background image")


!call fatal("Line 1068 displaying image.fits and tempimage.fits")

    image = tempImage
    deallocate(tempimage)
    nullify(tempimage)


end subroutine RefineBackgroundImage
!*****************************************************************
!
!
!****************************************************************
subroutine RefineBackgroundStrip(il, iu, jl, ju, image, refinedImage)
    integer(int16), intent(in) :: il, iu, jl, ju
    real(single), pointer :: image(:, :), refinedImage(:, :)

    real(single) :: globalMode, sigma 
    integer(int16) :: il1, iu1, jl1, ju1, nx, ny
    integer(int16) :: i, j
    integer(int16) :: dx
    integer(int32) :: n
    real(single), pointer, dimension(:) :: x=>null(), y=>null(), f=>null()

    dx = 200 / binx
    nx = ubound(image, 1)
    ny = ubound(image, 2)

    il1 = max(0_int16, il - dx)
    iu1 = min(nx, iu + dx)
    jl1 = max(0_int16, jl - dx)
    ju1 = min(ny, ju + dx)



    n = (iu1 - il1 + 1) * (ju1 - jl1 + 1)

    x => allocateSingle(1, n, "x")
    y => allocateSingle(1, n, "y")
    f => allocateSingle(1, n, "f")
  

    globalMode = ComputeBackgroundMode3(image(il1 : iu1, jl1 : ju1), sigma)

!    write(*, *) "Line 980 global mode=", globalMode, " sigma=", sigma
    globalMode = globalMode + max(1.0, 0.25 * sigma)


    n = 0
    do j = jl, ju
        do i = il, iu
            if (image(i, j) .lt. globalMode .and. image(i, j) .ge. 0.0) then
                n = n + 1
                x(n) = real(i)
                y(n) = real(j)
                f(n) = image(i, j)
            end if
        end do
    end do

    do j = jl, ju
        do i = il, iu
            if (image(i, j) .ge. globalMode) then
                refinedimage(i, j) = NeighbourInterpolate(real(i), real(j), n, x, y, f)
            end if
        end do
    end do
     
    deallocate(x, y, f)
    nullify(x, y, f)

 
end subroutine RefineBackgroundStrip



end module omskybackgroundimage_mod













