! -*-f90-*-
!
! Author:    jbakker
! Generated: Wed Aug  9 17:30:51 MEST 2002
! Info:      interface for GUI c++code
! Note:
!  * The GUI should always call "init()" first, and "finish()" last.
!  * In comments, I usually talk about methods. A method may be either a
!    function OR subroutine within a module
!  * Any variable starting with 'my' is global to this module
!  * Please read the comments below ;-)

module omsource_glue
use param
use sla
use omsource_old
use omsourceparameter_mod
use omsourceparam_mod
use omsourceregion_mod
use ommagnitude_mod
use omprogress_mod
use omphotometry_mod   ! In mssllib for doing aperture photometry

    implicit none
   
    type(SourceParameters), pointer :: mySources(:) => null()           ! The global source-list
    type(OmSourceRegion), pointer ::   mySourceRegions(:) => null() !   The globalbackground source-list

    type(SourceParameters), pointer :: myOldSources(:) => null()        ! The old global source-list


    type(SourceParameters), pointer :: myBackgroundSources(:) => null() ! The globalbackground source-list
    type(OmSourceRegion), pointer ::   myBackgroundRegions(:) => null() ! The globalbackground region-list


    type(SourceParameters), pointer :: myAddedSources(:) => null()           ! The sources that the 
                                                                            ! user has selected

    type(SourceParameters), pointer :: myAddedBackgroundSources(:) => null() ! The background sources 

                                                                            !that the user has lande
    real ::  mycalapertureradius
   
    integer :: myImageNx, myImageNy, myNoSources

    type(PhotometryParameters) :: myPhotometryParameters
 
contains

   ! Info: This is always the first function that is called by the GUI;
   !       in here we initialise/load/whatever




subroutine init(loaded)
    logical, intent(out) :: loaded
    
 
!write(*, *) "omsource_glue line 60 init"   
    call omsource_init(mySources, myBackgroundSources, &
                       myAddedSources, myAddedBackgroundSources, &
                       myPhotometryParameters, myimagenx, myimageny) 

    loaded = associated(mySources)

!write(*, *) "omsource_glue line 67 init loaded=", loaded   


    if(loaded) then
        myNoSources = size(mySources)
    else
        myNoSources = 0
    end if

!write(*, *) "omsource_glue line 76 myNoSources=", myNoSources   


end subroutine init

 


   ! Info: If the GUI closes, it calls
 
subroutine finish()
     
    call omsource_finish(mySources, myOldSources, myBackgroundSources, &
                       myAddedSources, myAddedBackgroundSources)
 
end subroutine finish


subroutine loadimage(image)
real(single), pointer :: image(:, :)

    image=>myImage
 
end subroutine loadimage

!*******************************************************************
!
!
!
!*******************************************************************

!***********************************************************************
!
!
!***********************************************************************
integer(int8) function getSourceExtensionFlag(sourceNumber)
    integer, intent(in) :: sourceNumber

    if ( .not. associated(mySources)) then
        getSourceExtensionFlag = 0 
        return
    end if


    if ( sourceNumber .gt.  size(mySources)) then
        getSourceExtensionFlag = 0 
        return
    end if

    getSourceExtensionFlag = mySources(sourceNumber)%extensionFlag


end function  getSourceExtensionFlag


!*************************************************************************
!
!
!
!**************************************************************************





! Subroutine for loading the sources from the source-list file
! produced by omdetect in omsourcebrower::SourceDetection
subroutine loadsources()
    character(500) :: fileName
    logical :: loaded = .false.


   



    fileName = trim(stringParameter("wdxset") )
    call omsource_loadsources(mySources, fileName)



    loaded = associated(mySources)




    if(loaded) then
        myNoSources = size(mySources)
    else
        myNoSources = 0
    end if




end subroutine loadsources





   ! Info: Recompute a single source as aperture, and background radii
   !       may have changed
subroutine recompute(id)
    integer, intent(in) :: id
      
!write(*, *) "omsource_glue recompute id=", id

    call omsource_recompute(mySources, id)
end subroutine recompute
        



!*************************************************************************
!
! Info: Remove a single source at location id
!*************************************************************************
subroutine remove(id)
    integer, intent(in) :: id
      



!write(*, *) "Line 310 omsource_glue line 310 removing mysources ", id

    call omsource_delete(mySources, id)



    if(associated(mySources) ) then
        myNoSources = size(mySources)
    end if




!write(*, *) "Line 310 omsource_glue line 310 removing myoldsources ", id



!    call omsource_delete(myOldSources, id)




end subroutine remove

! Info: deletes the whole source list
subroutine deleteList()
    call omsource_deleteList(mySources)


    myNoSources = 0
    


end subroutine deleteList


!**********************************************
! Replaces the new list with the old one
!*********************************************
subroutine restoreOldList()

    
    if (associated(mySources) .and. associated(myOldSources)) then
       

        call delete(mySources) 
        call new(mySources, low = 1, high = size(myOldSources))
 
      
        mySources = myOldSources

        if(associated(mySources) ) then
            myNoSources = size(mySources)
        end if





    end if
    
end subroutine restoreOldList

!**********************************************
! Replaces the new list with the old one
!*********************************************
subroutine saveOldList()

    if (associated(myOldSources)) then
        call delete(myOldSources)
    end if

    if (associated(mySources)) then
  
        call new(myOldSources, low = 1, high = size(mySources))

         myOldSources = mySources
     
 
    end if
    
   
end subroutine saveOldList

!********************************************
! Replaces a source
!*******************************************
subroutine replace(id, source)
  
    integer, intent(in) :: id
    type(SourceParameters), intent(in) :: source
   
    mySources(id) = source
    !write(*, *) "L293 omsource_glue replaceid=", id,&
    !     " source%rawbackgroundcountrate=", mySources(id)%rawbackgroundcountrate
end subroutine replace    


! Info: User has selected regions in the image; now it is up to us to:
!       - identify regions that are not found in the sourcelist already
!       - enlarge the sourcelist with information within the regions
!       - compute the source properties for the added sources
subroutine addRegions(regions, radius)
    type(OmSourceRegion),intent(in) :: regions(:)
    real(single), intent(in) :: radius

    call omsource_add(mySources, regions, radius, myAddedSources)
 
!    call SetTheSourceFlags(mySources)
 
  
end subroutine addRegions



subroutine addRegions1(regions)
    type(OmSourceRegion),intent(in) :: regions(:)
  
    call omsource_add1(mySources, regions, myAddedSources)
!    call SetTheSourceFlags(mySources)





    if(associated(mySources) ) then
        myNoSources = size(mySources)
    end if




   
end subroutine addRegions1


subroutine addRegions2(regions)
    type(OmSourceRegion),intent(in) :: regions(:)
  
    call omsource_add2(mySources, regions, myAddedSources)
!    call SetTheSourceFlags(mySources)

    if(associated(mySources) ) then
        myNoSources = size(mySources)
    end if


  
   
end subroutine addRegions2


subroutine centroidRegions()

    call omsource_computeparameters(myAddedSources)
   
end subroutine centroidregions





!************************************************************************
!
!
!************************************************************************

subroutine addBackgroundRegions(regions, radius)
    type(OmSourceRegion), intent(inout) :: regions(:)
    real(single), intent(in) :: radius
    call omsource_addBackgroundSources(mySources, myBackgroundSources, regions, radius)
end subroutine addBackgroundRegions



!***************************************************
!
!
!***************************************************
subroutine AppendToSourceRegions(regions)
    type(OmSourceRegion), intent(in) :: regions(:)
    type(OmSourceRegion), pointer :: tempRegions(:) => null()
    integer :: size1, size2 , size3   
    


    if (.not. associated(mySourceRegions)) then
        call new(mySourceRegions, 1, ubound(regions, 1))
        mySourceRegions = regions



        return
    end if
   


  
    size1 = ubound(regions, 1);
    if (size1 .eq. 0) return;
   
  
    size2 = ubound(mySourceRegions, 1)

    size3 = size1 + size2
    call new(tempRegions, 1, size2)
    

    tempRegions = mySourceRegions
    

    call delete(mySourceRegions)
    call new (mySourceRegions, 1, size3)



    mySourceRegions(1 : size2) = tempRegions
 
    call delete(tempRegions)

    ! The new background regions
    mySourceRegions(size2 + 1 : size3) = regions

  
end subroutine AppendToSourceRegions



!***************************************************
!
!
!***************************************************
subroutine AppendToBackgroundRegions(regions)
    type(OmSourceRegion), intent(in) :: regions(:)
    type(OmSourceRegion), pointer :: tempRegions(:) => null()
    integer :: size1, size2 , size3   
   


    if (.not. associated(myBackgroundRegions)) then
        call new(myBackgroundRegions, 1, ubound(regions, 1))
        myBackgroundRegions = regions



        return
    end if
   
  
    size1 = ubound(regions, 1);
    if (size1 .eq. 0) return;
   
  
    size2 = ubound(myBackgroundRegions, 1)

    size3 = size1 + size2
    call new(tempRegions, 1, size2)
    

    tempRegions = myBackgroundRegions
    

    call delete(myBackgroundRegions)
    call new (myBackgroundRegions, 1, size3)



    myBackgroundRegions(1 : size2) = tempRegions
 
    call delete(tempRegions)

    ! The new background regions
    myBackgroundRegions(size2 + 1 : size3) = regions

  
end subroutine AppendToBackgroundRegions
!***************************************************
!
!
!***************************************************

subroutine getBackgroundRegions(regions)
    type(OmSourceRegion),intent(in) :: regions(:)


    call omsource_getnewsources(mySources, regions, myBackgroundSources)

end subroutine getBackgroundRegions
!***********************************************
!
!********************************************** 

subroutine computebackground(value, sigma)
    real(single), intent(out) :: value, sigma

   value = backgroundMode(myImage, int(size(myImage, 1), int16), &
           int(size(myImage, 2), int16), 2.0, sigma) 
end subroutine computebackground


!********************************************
! GetApertureCounts
!*******************************************
subroutine GetTheCounts
   
    integer(int32) :: i      

    if (.not. associated(myBackgroundSources)) then
        return
    end if
    do i = 1, size(myBackgroundSources)
         myBackgroundSources(i)%exposure = 1.0
        call GetCounts1(myImage, myBackgroundSources(i))
    end do
end subroutine GetTheCounts    


!********************************************
! GetSourceCounts
! Calls GetCounts2 in omphotometry_mod in mssllib
!*******************************************
subroutine GetSourceCounts(x, y, radius, counts)
    real(single), intent(in) :: x, y, radius
    real(single), intent(out) :: counts
        

   
    call GetCounts2(myImage, x, y, radius, counts)
    
end subroutine GetSourceCounts    

!********************************************
! GetImageCounts
! Calls GetCounts2 in omphotometry_mod in mssllib
!*******************************************
real(single) Function GetImageCounts(xPos, yPos, radius)
    real(single), intent(in) :: xPos, yPos, radius

    if (.not. associated(myImage)) call fatal("myImage is not associted", "GetImageCounts")

    call GetCounts2(myImage, xPos, yPos, radius, GetImageCounts)

end Function GetImageCounts  

!********************************************
!
! GetBackgroundCounts
!
!*******************************************
subroutine GetTheBackgroundCounts(x, y, radius, r1, r2in, counts)
    real(single), intent(in) :: x, y, radius, r1, r2in
    real(single), intent(out) :: counts
    real(single) :: r2

    r2 = r2In
   
    call GetBackgroundCounts(myImage, x, y, radius, r1, r2, counts)
    
end subroutine GetTheBackgroundCounts    

!********************************************
! deleteBackgroundList
!*******************************************
subroutine deleteBackgroundList()
!write(*, *) "omsource_glue line 459 deleting background list"   
    call omsource_deleteList(myBackgroundSources)
    
end subroutine deleteBackgroundList




real(single) Function GetPSFFraction(xPos, yPos, radius)
    real(single), intent(in) :: xPos, yPos, radius

    GetPSFFraction = PSFFractionOfTotalCounts(&
              xPos, yPos, radius, 0.5)

end Function GetPSFFraction  

!*******************************************************
!
!******************************************************
real(single) Function GetTheBackgroundMode()

    GetTheBackgroundMode = omsource_GetBackgroundMode()

end Function GetTheBackgroundMode
!********************************************
! DeleteBackgroundRegions
!*******************************************
subroutine DeleteBackgroundRegions
    
!write(*, *) "omsourceglue line 488 deletebackgroundregions"
    call delete(myBackgroundRegions)
  

end subroutine DeleteBackgroundRegions
!******************************************************
!
!
!*****************************************************

subroutine InitialisePhotometry

    call InitialiseArrays

end subroutine InitialisePhotometry
subroutine Test1(n, data)
    integer, intent(in) :: n
    real data(:)
    integer i

!write(*, *) "Test1"

!    do i = 1, n
!        write(*, *) i, data(i)
!    end do
    
end subroutine Test1


subroutine ComputeSourceParameters(sources)
   type(SourceParameters), pointer :: sources(:)
 

   call omsource_computeparameters(sources)
end subroutine ComputeSourceParameters


subroutine ComputeSourcePositions(sources)
   type(SourceParameters), pointer :: sources(:)
 

   call omsource_computesourcepositions(sources)
end subroutine ComputeSourcePositions



!************************************************************
!
!***********************************************************
subroutine ComputeSourcePositions2( source )
    type(SourceParameters) :: source
    type(SourceParameters), pointer :: sources(:)

    allocate(sources(1))

   
    sources(1) = source
    call computesourcepositions(sources) 
   

    source = sources(1)



    deallocate(sources)
    nullify(sources)
end subroutine ComputeSourcePositions2





!************************************************************
!
!***********************************************************
subroutine ComputePhotometry(sources)
    type(SourceParameters), pointer :: sources(:)
    !write(*,*) "L670 ComputePhotometry"
    call omsource_photometry(myImage, size(sources), sources) 
 
end subroutine ComputePhotometry

!************************************************************
!
!***********************************************************
subroutine ComputePhotometry2( source )
    type(SourceParameters) :: source
    type(SourceParameters), pointer :: sources(:)
    !write(*,*) "L681 ComputePhotometry2"
    allocate(sources(1))

!    write(*, *) "omsource-glue Compute photometry2 line 67 call omsource_photometry"
    sources(1) = source
    call omsource_photometry(myImage, 1, sources) 
   
    source = sources(1)

!write(*, *) "omsource_glue line 573 mag=", source%magnitude, " +/-", source%magnitudeerr

    deallocate(sources)
    nullify(sources)
end subroutine ComputePhotometry2















!*****************************************************
!
!****************************************************

subroutine AppendToSourceList(newSources)
   type(SourceParameters), pointer :: newSources(:)
   type(SourceParameters), pointer :: tempSources(:)
   integer(int32) :: k

   if (.not. associated(newSources)) return; 
   if (.not. associated(mySources)) then
        mySources => newSources
        nullify(newSources)
        return
   end if

   
   k = size(mySources)
 
   call new(tempSources, low = 1, high = k)
 
   tempSources = mySources
   deallocate(mySources)
   nullify(mySources)
       
   call new(mySources, low = 1, high = k + size(newSources))
 
   mySources(1 : k) = tempSources
   deallocate(tempSources)
   nullify(tempSources)
   mySources(k + 1 : k + size(newSources)) = newSources

   myNoSources = size(mySources)

     
end subroutine AppendToSourceList


!*******************************************
! Function ComputeEpoch
!******************************************
real(double) function  ComputeEpoch(dateObs)
   
    character(19), intent(in)  :: dateObs               
    real(double)  :: JulianDate0, JulianDate
    character(19)  :: missionReferenceTime = "1998-01-01T00:00:00"
    
    ! Compute the Julian date of the mission reference time.
    JulianDate0 = ModifiedJulianDate(missionReferenceTime)

    ! Compute the Julian date of the observation
    JulianDate = ModifiedJulianDate(dateObs)

    
    ComputeEpoch = 1998.0_double + (JulianDate - JulianDate0) / 365.25_double

end function ComputeEpoch


!*************************************************************
!
!*************************************************************
subroutine ComputeJulianDates(JDStart, JDEnd, JDMiddle)
    real(double), intent(out) :: JDStart, JDEnd, JDMiddle
    type(DataSetT) :: dataSetH
    character(20) :: string
    
    dataSetH = DataSet(trim(stringParameter("imageset")), read)


    string = trim(stringAttribute(dataSetH, "DATE-OBS"))

    JDStart = ModifiedJulianDate(string)

    string = trim(stringAttribute(dataSetH, "DATE-END"))

    JDEnd = ModifiedJulianDate(string)

 

    call release(dataSetH)


    JDMiddle = (JDStart + JDEnd) / 2.0_double

end subroutine ComputeJulianDates
!****************************************************************************************
!
!***************************************************************************************

subroutine CreateTheRegionFile()
! integer i



if ( .not. associated(mySources)) return

!write(*, *) "craetetheregionfile, wdxset=", trim(stringParameter("wdxset"))


!write(*, *) "associated mysources=", associated(mySources)

!write(*, *) "size mysources=" , size(mysources)

!do i = 1, size(mySources)

!write(*, *) i, ") x= ", mysources(i)%xpos, " y=", mysources(i)%ypos, "a=", mysources(i)%semia, &
!" b=", mysources(i)%semib, " theta=", mySources(i)%theta, " qflag=", mySources(i)%qualityFlag
!end do
     
    call createRegionFile1(size(mysources), mysources, trim(stringParameter("wdxset") ))
!write(*, *) "Line 642"

end subroutine CreateTheRegionFile


!****************************************************************************************
!
!***************************************************************************************

subroutine computeBackgroundImage
 
    integer(int16) :: dx, nx, ny

return
    nx = size(myImage, 1)

    ny = size(myImage, 2)


    dx = 50 / myImageWindow%bx
    if (dx .gt. nx) dx = max(nx, 20_int16)
    if (dx .gt. ny) dx = max(ny, 20_int16)

!write(*, *) "omsourceold line 679 creating background image"


!    call createBackgroundImage(myimage, myQuality, dx, &
!                    mybackgroundimage, mybackgroundsigmaimage, maxheight = 10000000.0) ! ws 10000

!write(*, *) "omsourceold line 684 finished creating background image"



end subroutine computeBackgroundImage


!*******************************************************************************
!
!*******************************************************************************

subroutine getoldsource( no, oldSource )
    integer, intent(in) :: no
    type(SourceParameters), intent(out) :: oldSource


    oldsource = mysources(no)
    

end subroutine getOldSource


!*******************************************************************************
!
!*******************************************************************************

subroutine getnewsource( no, newSource )
    integer, intent(in) :: no
    type(SourceParameters), intent(out) :: newSource


    newsource = mysources(no)
    

end subroutine getnewSource





!*******************************************************************************
!
!*******************************************************************************
subroutine recompute_photometry(numberToFit, no, type, radius, &
      innerA, outerA, axialRatio, theta, background)


    integer, intent(in) :: numberToFit
    integer, intent(in) :: no(1 : numberToFit)

    
    integer, intent(in) :: type
    real(double), intent(in) :: radius, innerA, outerA, axialRatio, theta, background

!write(*, *) "omsource_glue line 8181 recompute_photometry source numbers=", &
!no(1 : numberToFit), " numbertofit=", numberToFit 



    !******************************************************************************
    ! The module is in omsource_old.f90
    !*****************************************************************************
    
    call omsourceold_recomputephotometry(mySources, numberToFit, no, type, radius, &
                                innerA, outerA, axialRatio, theta, background)

end subroutine recompute_photometry


end module omsource_glue














