
! -*-f90-*-
!
! Author:    jbakker
! Generated: Mon Sep 23 13:17:55 CEST 2002
! Usage:     -
! Info:      This is a **temporary** module and acts as a placeholder
!            until all SAS fortran developers have sorted out their act
!            together
module omsourceutils_mod
   use types
   use errorHandling, only : fatal
   implicit none

   public ! all

   ! useful global parameters

contains

   ! Info: Returns an sorted array
   ! Example:
   !     real(single)   myUnsortedArray(20)
   !     real(single)   mySortedArray(20)
   !     :
   !     :
   !     mySortedArray=sorted(myUnsortedArray)
   function sorted(unsortedArray) result(sortedArray)
      real(single),  intent(in)          :: unsortedArray(:)
      real(single)                       :: sortedArray(size(unsortedArray))

      sortedArray=unsortedArray(sortedIndexOf(unsortedArray))
   end function sorted



   ! Info: Returns an array of sorted indices.
   ! Example:
   !     real(single)   myArray(20)
   !     integer(int32) myIndex(20)
   !     :
   !     :
   !     myIndex=sortedIndexOf(myArray)
   function sortedIndexOf(array) result(sortedIndices)
      real(single),  intent(in)          :: array(:)
      integer(int32)                     :: sortedIndices(size(array))

      integer(int32)                     :: sizeArray
      integer(int32)                     :: i, ir, j, l, irra
      real(single)                       :: rra

      sizeArray=size(array)

      ! immediate return if:
      if (sizeArray == 0) return
      if (sizeArray < 2) then
         sortedIndices(1) = 1
         return
      end if

      ! initial
      do i = 1, sizeArray
         sortedIndices(i) = i
      end do

      l = sizeArray / 2 + 1
      ir = sizeArray   

      do 
         if (l > 1) then
            l = l - 1
            rra = array(sortedIndices(l))   
            irra = sortedIndices(l)    
         else
            rra = array(sortedIndices(ir))        
            irra = sortedIndices(ir)

            sortedIndices(ir) = sortedIndices(1)
            ir = ir - 1

            if (ir <= 1) then

               sortedIndices(1) = irra
               exit
            end if
         end if

         i = l
         j = l + l
         do while(j <= ir)

            if (j < ir) then
               if (array(sortedIndices(j)) < array(sortedIndices(j+1))) j=j+1
            end if
            if (rra < array(sortedIndices(j))) then

               sortedIndices(i) = sortedIndices(j)
               i = j
               j = j + j
            else
               exit
            end if

         end do

         sortedIndices(i) = irra
      end do
   end function sortedIndexOf



   ! Info: Returns the median of an unsorted array
   ! Example:
   !     real(single)   myArray(20)
   !     real(single)   median
   !     :
   !     :
   !     median=medianOfUnsorted(myArray)
   function medianOfUnsorted(array) result(median)
      real(single), intent(in) :: array(:)
      real(single)             :: median

      median=medianOfSorted(array(sortedIndexOf(array)))
   end function medianOfUnsorted



   ! Info: Returns the median of a sorted array
   ! Example:
   !     real(single)   myArray(20)
   !     real(single)   median
   !     :
   !     :
   !     median=medianOfSorted(myArray)
   function medianOfSorted(array) result(median)
      real(single), intent(in) :: array(:)
      real(single)             :: median

      integer(int32)           :: sizeArray
      integer(int32)           :: k


      sizeArray=size(array)

      ! immediate return if:
      if (sizeArray == 0) then 
         median=0.0
         return
      end if
      if (sizeArray == 1) then 
         median = array(1)
         return
      end if

      k = sizeArray / 2
      if (k * 2 .eq. sizeArray) then   ! Even number
         median  = (array(k) + array(k + 1)) / 2.0
      else
         median = array(k + 1)
      end if
   end function medianOfSorted



   ! Info: Returns the quadrant of 2-dimensional coordinates!
   !       The result is in range [0, 4], where 0 denotes the coordinate
   !       is on axis...
   ! Example:
   !       call assertOn(associated(ptr),"Memory leak?")
   !       deallocate(ptr)
   !       nullify(ptr)
   subroutine assertOn(statement,comment)
      logical,     intent(in) :: statement
      character(*),intent(in) :: comment
      if (statement) then
         call fatal("assertOn",comment)
      end if
   end subroutine assertOn


end module omsourceutils_mod


